#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>

#include "modsetup.h"

/* v1.5 split program by single modules */
#include "xscommon.h"
#include "xsmodmsk.h"
#include "xsmodvoc.h"
#include "xsmodsvg.h"
#include "xsmodstr.h"
#include "xsmodmap.h"
#include "textinfo.h"

#define GAMESCOUNT 2
#define INFOSCOUNT 14
#define FIXESCOUNT 100
#define GET_OFFS(x) ((x) & 0x00FFFFFFUL)

#pragma pack(push, 1)
typedef struct {
  DWORD offs;
  BYTE  data;
} fix_item;

typedef struct {
  char *name;
  char *game;
  DWORD size;
} fix_list;

typedef struct {
  BYTE idx;
  BYTE count;
  BYTE stat;
} fix_info;
#pragma pack(pop)

static fix_list gamelist[GAMESCOUNT] = {
  {"EX.EXE", "Exhumed",    825341UL},
  {"PS.EXE", "PowerSlave", 818966UL}
};

static fix_info infolist[INFOSCOUNT] = {
  /* Revert [Run] logic (always run) */
  { 0,  1, 0},
  /* Always [Mouseview] */
  { 1,  4, 0},
  /* v1.5 Tweak strafe buttons code */
  { 5, 16, 0},
  /* /CDOFF always enabled (No-CD) */
  {21,  2, 0},
  /* Allow F12 key to take screenshots */
  {23,  8, 0},
  /* Fix C key to work properly */
  {31,  4, 0},
  /* Fix incorrect cheat codes messages */
  {35, 22, 0},
  /* Invert mouse aim */
  {57,  2, 0},
  /* Change throw grenade routine */
  {59,  4, 0},
  /* Disable replaying level checkpoint load */
  {63,  1, 0},
  /* Disable player movement prediction for enemies */
  {64,  1, 0},
  /* Show subtitles */
  {65,  1, 0},
  /* Disable Omen Wasps flying sound */
  {66,  1, 0},
  /* Custom BUILD screen resolutions */
  {67, 33, 0}
};

/* cheat text fix */
#define CHT_IDX 6
#define CHT_LEN 4
/* v1.5 sfx hack */
#define VOC_IDX 12
/* screen resolution hack item index */
#define SCR_IDX 13

static fix_item datalist[FIXESCOUNT][GAMESCOUNT] = {
/*
  packed data types
  {{0xAAOOOOOO, 0xBB}, {0xOOOOOOOO, 0xBB}, ...}
  AA - old byte
  BB - new byte
  OOOOOO - offset
*/
  /* Run (7) - revert logic */
  {{0x740367FEUL, 0x75}, {0x7403639BUL, 0x75}}, /* always run */
  /* Mouseview (25) - always on */
  {{0xE9036887UL, 0xB8}, {0xE9036424UL, 0xB8}}, /* [3] allow mouseview when Move_Forward */
  {{0xE90368B1UL, 0xBA}, {0xE903644EUL, 0xBA}}, /* [4] allow mouseview when Move_Backward */
  {{0x420368EFUL, 0x00}, {0x4203648CUL, 0x00}}, /* [1] always mouseview */
  {{0x02039A7EUL, 0x00}, {0x020395B9UL, 0x00}}, /* [2] turn off center view */
  /* v1.5 fix strafe buttons */
  /* # PlayerInterruptKeys */
  /* test 5th bit - Strafe_Left */
  {{0x890369D9UL, 0xA8}, {0x89036576UL, 0xA8}},
  {{0xCA0369DAUL, 0x20}, {0xCA036577UL, 0x20}},
  /* jump to jz code (keep registers unmodified) */
  {{0xC10369DBUL, 0xEB}, {0xC1036578UL, 0xEB}},
  {{0xE80369DCUL, 0x28}, {0xE8036579UL, 0x28}},
  /* lPlayerXVel += addX [addX == ebx] */
  {{0x8B036A07UL, 0x01}, {0x8B0365A4UL, 0x01}},
  {{0x0D036A08UL, 0x1D}, {0x0D0365A5UL, 0x1D}},
  /* lPlayerYVel -= addY [addY == ecx] */
  {{0x8B036A0DUL, 0x29}, {0x8B0365AAUL, 0x29}},
  {{0x3D036A0EUL, 0x0D}, {0x3D0365ABUL, 0x0D}},
  /* jump to next code block */
  {{0x01036A13UL, 0xEB}, {0x010365B0UL, 0xEB}},
  {{0xC1036A14UL, 0x0D}, {0xC10365B1UL, 0x0D}},
  /* Strafe_Right */
  /* lPlayerXVel -= addX [addX == ebx] */
  {{0x8B036A30UL, 0x29}, {0x8B0365CDUL, 0x29}},
  {{0x15036A31UL, 0x1D}, {0x150365CEUL, 0x1D}},
  /* lPlayerYVel += addY [addY == ecx] */
  {{0x8B036A36UL, 0x01}, {0x8B0365D3UL, 0x01}},
  {{0x1D036A37UL, 0x0D}, {0x1D0365D4UL, 0x0D}},
  /* jump to next code block */
  {{0x29036A3CUL, 0xEB}, {0x290365D9UL, 0xEB}},
  {{0xC2036A3DUL, 0x13}, {0xC20365DAUL, 0x13}},
  /* /cdoff always enabled */
  {{0x000A2EC2UL, 0x01}, {0x000A1B8AUL, 0x01}}, /* No-CD */
  /* that patch byte needed only for Exhumed - skips MSCDEX check */
  {{0x7505AB77UL, 0xEB}, {0x00000000UL, 0x00}},
  /* v1.2 - screenshots and PC Speaker patch */
  /* disable PC Speaker (A LOT faster screenshot save) */
  {{0xBA066092UL, 0xEB}, {0xBA065B84UL, 0xEB}},
  {{0x61066093UL, 0x07}, {0x61065B85UL, 0x07}},
  {{0xBA0661ECUL, 0xEB}, {0xBA065CDEUL, 0xEB}},
  {{0x610661EDUL, 0x0A}, {0x61065CDFUL, 0x0A}},
  /* EX: cmp b,[00273AC9], 000 -> test b,[00273EB8], 000 */
  /* PS: cmp b,[00273AF5], 000 -> test b,[00273EE4], 000 */
  /* relocation table */
  {{0xC90046ECUL, 0xB8}, {0xF5004762UL, 0xE4}},
  {{0xC90047BBUL, 0xB8}, {0xF5004772UL, 0xE4}},
  {{0xC902FC36UL, 0xB8}, {0xF502F792UL, 0xE4}},
  {{0xC902FC4DUL, 0xB8}, {0xF502F7A9UL, 0xE4}},
  /* # CheckKeys */
  /* v1.2 - fix C key to work properly + v1.4 rework */
  {{0x8402DFEAUL, 0xD0}, {0x8402DBEBUL, 0xD0}},
  {{0xDB02DFEBUL, 0xEB}, {0xDB02DBECUL, 0xEB}},
  {{0x7402DFECUL, 0x73}, {0x7402DBEDUL, 0x73}},
  {{0x2702DFF6UL, 0x34}, {0x2702DBF7UL, 0x34}},
  /* v1.3 cheats messages fixes */
  /* fix incorrect message "SNAKE CAM ENABLED" -> "FLASHES TOGGLED" for only not implemented cheat */
  {{0x2502C929UL, 0x37}, {0x3002C738UL, 0x42}},
  {{0x250A73C4UL, 0x37}, {0x300A6174UL, 0x42}},
  /* remove incorrect message "FLASHES TOGGLED" for keys cheat (visible due few frames) */
  {{0x3702C922UL, 0xF9}, {0x4202C731UL, 0xF1}},
  {{0x2E02C923UL, 0x26}, {0x3002C732UL, 0x26}},
  {{0x370A73C8UL, 0xF9}, {0x420A6178UL, 0xF1}},
  {{0x2E0A73C9UL, 0x26}, {0x300A6179UL, 0x26}},
  /* remove incorrect message "FULL MAP" for clip cheat (visible due few frames) */
  {{0x4702C91BUL, 0xF9}, {0x5202C72AUL, 0xF1}},
  {{0x2E02C91CUL, 0x26}, {0x3002C72BUL, 0x26}},
  {{0x470A73CCUL, 0xF9}, {0x520A617CUL, 0xF1}},
  {{0x2E0A73CDUL, 0x26}, {0x300A617DUL, 0x26}},
  /* restore message "SNAKE CAM ENABLED" for cobra cheat */
  {{0xF902C914UL, 0x25}, {0xF102C723UL, 0x30}},
  {{0x2602C915UL, 0x2E}, {0x2602C724UL, 0x30}},
  {{0xF90A73D0UL, 0x25}, {0xF10A6180UL, 0x30}},
  {{0x260A73D1UL, 0x2E}, {0x260A6181UL, 0x30}},
  /* restore message "FULL MAP" for map cheat */
  {{0xF902C90DUL, 0x47}, {0xF102C71CUL, 0x52}},
  {{0x2602C90EUL, 0x2E}, {0x2602C71DUL, 0x30}},
  {{0xF90A73D4UL, 0x47}, {0xF10A6184UL, 0x52}},
  {{0x260A73D5UL, 0x2E}, {0x260A6185UL, 0x30}},
  /* fix cobra cheat message "ENABLED" -> "TOGGLED" because it actually toggled */
  {{0xBA09F82FUL, 0xAB}, {0xBA09E63AUL, 0xAB}},
  {{0xB109F830UL, 0xB0}, {0xB109E63BUL, 0xB0}},
  {{0xBE09F831UL, 0xB8}, {0xBE09E63CUL, 0xB8}},
  {{0xBD09F832UL, 0xB8}, {0xBD09E63DUL, 0xB8}},
  /* # PlayerInterruptKeys */
  /* v1.4 invert mouse */
  /* sub eax, edx -> add eax, edx */
  {{0x29036904UL, 0x03}, {0x290364A1UL, 0x03}},
  {{0xD0036905UL, 0xC2}, {0xD00364A2UL, 0xC2}},
  /* # ThrowGrenade */
  /* v1.4 remove total velocity from the grenade throw code */
  /* only elapsed time used (0x5A - word_1B831E[ebx]) */
  {{0x0F04A112UL, 0x8B}, {0x0F049C44UL, 0x8B}},
  {{0xBF04A113UL, 0xD1}, {0xBF049C45UL, 0xD1}},
  /* skip _totalvel */
  {{0x5404A114UL, 0xEB}, {0x54049C46UL, 0xEB}},
  {{0x2404A115UL, 0x08}, {0x24049C47UL, 0x08}},
  /* # BuildBullet */
  /* v1.5 disable player movement prediction for enemies (bugged) */
  {{0x7E048546UL, 0xEB}, {0x7E048078UL, 0xEB}},
  /* # main */
  /* v1.5 do not load the last checkpoint when replaying any level twice in a row */
  {{0x7502F609UL, 0xEB}, {0x7502F183UL, 0xEB}},
  /* # InitSpiritHead */
  /* v1.5 always show subtitles */
  {{0xA3030A42UL, 0xA1}, {0xA30305ADUL, 0xA1}},
  /* # LoadSound */
  /* v1.5 disable annoying Omen Wasps flying sfx */
  /* force kopen4load() to load .VOC files from the current folder) */
  {{0x01045E59UL, 0x00}, {0x0104598BUL, 0x00}},
  /* v1.1 - vid_hack */
  /* always hi-res mode as "ScreenWidth = 640" in SETUP.CFG */
  {{0x05049D64UL, 0x00}, {0x050499B7UL, 0x00}},
  /* read "STATUS.MSK" from current directory instead of STUFF.DAT */
  {{0x0105DC46UL, 0x00}, {0x0105D73BUL, 0x00}},
  /* disable x2 scale for hires */
  {{0x7405DCFCUL, 0xEB}, {0x7405D7F1UL, 0xEB}},
  {{0x8002EA6CUL, 0x80}, {0x8002E5ECUL, 0x80}},
  /* width */
  {{0x0202EA6DUL, 0x02}, {0x0202E5EDUL, 0x02}},
  {{0x9002EA71UL, 0x90}, {0x9002E5F1UL, 0x90}},
  /* height */
  {{0x0102EA72UL, 0x01}, {0x0102E5F2UL, 0x01}},
  {{0x0C05DC00UL, 0x8C}, {0x0C05D6F5UL, 0x8C}},
  /* v1.3 increase stack size for hi-res buffer (1024*2) + 12 => (1600*2) + 12 */
  {{0x0805DC01UL, 0x0C}, {0x0805D6F6UL, 0x0C}},
  {{0x0005DC09UL, 0x80}, {0x0005D6FEUL, 0x80}},
  {{0x0805DC0AUL, 0x0C}, {0x0805D6FFUL, 0x0C}},
  {{0x0005DC12UL, 0x80}, {0x0005D707UL, 0x80}},
  {{0x0805DC13UL, 0x0C}, {0x0805D708UL, 0x0C}},
  {{0x0005DC23UL, 0x80}, {0x0005D718UL, 0x80}},
  {{0x0805DC24UL, 0x0C}, {0x0805D719UL, 0x0C}},
  {{0x0005DC32UL, 0x80}, {0x0005D727UL, 0x80}},
  {{0x0805DC33UL, 0x0C}, {0x0805D728UL, 0x0C}},
  {{0x0405DC6CUL, 0x84}, {0x0405D761UL, 0x84}},
  {{0x0805DC6DUL, 0x0C}, {0x0805D762UL, 0x0C}},
  {{0x0405DC79UL, 0x84}, {0x0405D76EUL, 0x84}},
  {{0x0805DC7AUL, 0x0C}, {0x0805D76FUL, 0x0C}},
  {{0x0005DC87UL, 0x80}, {0x0005D77CUL, 0x80}},
  {{0x0805DC88UL, 0x0C}, {0x0805D77DUL, 0x0C}},
  {{0x0805DC8EUL, 0x88}, {0x0805D783UL, 0x88}},
  {{0x0805DC8FUL, 0x0C}, {0x0805D784UL, 0x0C}},
  {{0x0405DC9FUL, 0x84}, {0x0405D794UL, 0x84}},
  {{0x0805DCA0UL, 0x0C}, {0x0805D795UL, 0x0C}},
  {{0x0005DCBAUL, 0x80}, {0x0005D7AFUL, 0x80}},
  {{0x0805DCBBUL, 0x0C}, {0x0805D7B0UL, 0x0C}},
  {{0x0805DD36UL, 0x88}, {0x0805D82BUL, 0x88}},
  {{0x0805DD37UL, 0x0C}, {0x0805D82CUL, 0x0C}},
  {{0x0C05DD40UL, 0x8C}, {0x0C05D835UL, 0x8C}},
  {{0x0805DD41UL, 0x0C}, {0x0805D836UL, 0x0C}}
};

/* v1.5 */
#define MODSLCOUNT 4
static fix_info modsinfo[MODSLCOUNT] = {
  /* Skip game title */
  {0, 1, 0},
  /* Change automapping type */
  {1, 2, 0},
  /* Show invisible objects */
  {3, 2, 0},
  /* Allow XMIDI instead of CD music */
  {5, 7, 0}
};

#define MODSCOUNT 12
static fix_item modslist[MODSCOUNT][GAMESCOUNT] = {
  /* Skip game title */
  /* #main */
  /* test si, si => xor si, si */
  {{0x8502F1F5UL, 0x33}, {0x8502ED7EUL, 0x33}},
  /* Change automapping type */
  /* # initengine */
  /* mov [...], bh (0) => inc b, [...] ; automapping = 1 */
  {{0x88064E4EUL, 0xFE}, {0x88064940UL, 0xFE}},
  {{0x3D064E4FUL, 0x05}, {0x3D064941UL, 0x05}},
  /* Show invisible objects */
  /* # initengine */
  /* mov [...], dh (0) => inc b, [...] ; showinvisibility = 1 */
  {{0x88064D4CUL, 0xFE}, {0x8806483EUL, 0xFE}},
  {{0x35064D4DUL, 0x05}, {0x3506483FUL, 0x05}},
  /* Allow XMIDI instead of CD music */
  /* # playCDtrack - disable CD music */
  {{0x0A05AE6CUL, 0x00}, {0x0A05A961UL, 0x00}},
  /* # InitFX => # InitMusic */
  /* jump to jump below from audio failed */
  {{0x5A045BE3UL, 0xEB}, {0x5A045715UL, 0xEB}},
  {{0x59045BE4UL, 0x21}, {0x59045716UL, 0x21}},
  /* jump from audio done */
  {{0x5A045C06UL, 0xEB}, {0x5A045738UL, 0xEB}},
  {{0x59045C07UL, 0x11}, {0x59045739UL, 0x11}},
  /* # UnIniFX => # UnInitMusic */
  {{0x5A045C12UL, 0xEB}, {0x5A045744UL, 0xEB}},
  {{0x59045C13UL, 0x56}, {0x59045745UL, 0x56}}
};

/* v1.5 old strafe fix to remove (bad implementation) */
#define OLD_COUNT 4
static fix_item old_list[OLD_COUNT][GAMESCOUNT] = {
  /* Strafe (4) - revert logic (always strafe; faster than strafe left / strafe right) */
  {{0x84036A81UL, 0x85}, {0x8403661EUL, 0x85}}, /* [1] revert logic */
  {{0x87036AA6UL, 0x57}, {0x87036643UL, 0x57}}, /* [2] go to mouse_look_turn */
  {{0x21036AB0UL, 0x1C}, {0x2103664DUL, 0x1C}}, /* [4] go to [3] */
  {{0x5F036ACEUL, 0x2F}, {0x5F03666BUL, 0x2F}}  /* [3] go to mouse_look_turn */
};

static menuitem maingame[GAMESCOUNT + 1];

static menuitem mainmenu[] = {
  {
    /* menu title */
    "Main menu",
    /* subheader (green) */
    NULL,
    /* total menu items (this not included) */
    7
  },
  {
    /* menu item name */
    "Hacks setup",
    /* menu item hint */
    "Configure various hacks.",
    /* menu item flag */
    0
  },
  {
    "Custom resolution",
    "Allow to set custom BUILD screen resolution.",
    0
  },
  {
    "Modding setup",
    "Advanced modifications (not recommended for casual playthrough).",
    0
  },
  {
    "Saved games editor",
    "Edit various parameters in saved games.",
    0
  },
  {
    "About DeHacker (+notes)",
    "Information about this program.",
    0
  },
  {
    "Save settings and exit",
    "Save the current settings and exit.",
    0
  },
  {
    "Exit without saving settings",
    "Exit without saving any changes.",
    0
  }
};

static menuitem mainctrl[] = {
  {"Hacks menu", NULL, 13},
  {
    "Revert [Run] logic (always run)",
    "Always run when [Run] key NOT pressed.",
    ITEM_CHKBOX
  },
  {
    "Always [Mouseview]",
    "Mouse look always enabled without blocking when moving forward / backward.",
    ITEM_CHKBOX
  },
  {
    "Tweak strafe buttons code",
    "Movement with [Strafe_Left] / [Strafe_Right] will be fast as forward / backward.",
    ITEM_CHKBOX
  },
  {
    "/CDOFF always enabled (No-CD)",
    "The game will work without CD. This also disables MSCDEX check for Exhumed.",
    ITEM_CHKBOX
  },
  {
    "Allow F12 key to take screenshots",
    "Enables native screenshot support and disables PC Speaker noises (FAST save).",
    ITEM_CHKBOX
  },
  {
    "Fix C key to work properly",
    "Stabilizes C key so it can be used for playing (Alt+C implementation bug).",
    ITEM_CHKBOX
  },
  {
    "Fix incorrect cheat codes messages",
    "Restores and repairs correct information messages for the cheat codes.",
    ITEM_CHKBOX
  },
  {
    "Invert mouse aim",
    "Invert vertical mouse movement (on the Y-axis).",
    ITEM_CHKBOX
  },
  {
    "Tweak grenade throw code",
    "Only [Fire] holding delay will affect distance (total player velocity excluded).",
    ITEM_CHKBOX
  },
  {
    "Fix replaying level checkpoint load",
    "Do not load the last checkpoint when replaying any level twice in a row.",
    ITEM_CHKBOX
  },
  {
    "Tweak player movement prediction for enemies",
    "Exclude player movement delta from prediction formula (fixes enemies shooting).",
    ITEM_CHKBOX
  },
  {
    "Always show subtitles in briefings",
    "Subtitles will be shown along the speech animation (end of level 1 and 11).",
    ITEM_CHKBOX
  },
  {
    "Disable Omen Wasps flying sound",
    "Replace annoying sound with empty file (note that enemy will approach silently).",
    ITEM_CHKBOX
  }
};

static menuitem mainwarn[] = {
  {"Please DO NOT use these options for casual playthrough!", NULL, 2 | MENU_DLGBOX},
  {"Cancel and return", "Return back to main menu.", 0},
  {"Yes, I understand", "Continue to modding menu.", 0}
};

static menuitem mainmods[] = {
  {"Modding menu", NULL, 10},
  {
    "Skip game title",
    "Skips title screens with intro movie.",
    ITEM_CHKBOX
  },
  {
    "Change automapping type",
    "Show only visible walls on the map instead of walls of the visited sectors.",
    ITEM_CHKBOX
  },
  {
    "Show invisible objects",
    "Show various hidden sprites and highlight current target with flying marker.",
    ITEM_CHKBOX
  },
  {
    "Allow XMIDI instead of CD music",
    "Requires *.MDI and LEV*.XMI (for each level) from Beta, set music card in setup.",
    ITEM_CHKBOX
  },
  {
    "Decrypt global strings block",
    "Briefings, cheats, item names and other strings will be in plain text.",
    ITEM_CHKBOX
  },
  {"", NULL, ITEM_GRAYED},
  {
    "Do not change world map data",
    "Do not apply any changes to world map data.",
    ITEM_RADBOX | ITEM_HILITE
  },
  {
    "Reset world map data to defaults",
    "Reset and restore world map data to game default values.",
    ITEM_RADBOX
  },
  {
    "Save world map data to file",
    "Save current world map data to \"WORLDMAP.INI\" file.",
    ITEM_RADBOX
  },
  {
    "Load world map data from file",
    "Load and replace world map data from \"WORLDMAP.INI\" file.",
    ITEM_RADBOX
  }
};

static menuitem mainmode[] = {
  {"Screen resolution menu", NULL, 14},
  {"Disabled",  "Disables hi-res hack and revert all changes back.", ITEM_HILITE | ITEM_RADBOX},
  { "320x200",  NULL, ITEM_RADBOX},
  { "360x200",  NULL, ITEM_RADBOX},
  { "320x240",  NULL, ITEM_RADBOX},
  { "360x240",  NULL, ITEM_RADBOX},
  { "320x400",  NULL, ITEM_RADBOX},
  { "360x400",  NULL, ITEM_RADBOX},
  { "640x350",  NULL, ITEM_RADBOX},
  { "640x400",  NULL, ITEM_RADBOX},
  { "640x480",  NULL, ITEM_RADBOX},
  { "800x600",  NULL, ITEM_RADBOX},
  {"1024x768",  NULL, ITEM_RADBOX},
  {"1280x1024", NULL, ITEM_RADBOX},
  {"1600x1200", NULL, ITEM_RADBOX}
};

static menuitem maininfo[] = {
  {"About DeHacker", NULL, 16 | MENU_NFOBOX},
  {"Exhumed / PowerSlave DeHacker Version 1.5", NULL, 0},
  {"(c) -=CHE@TER=- 2014,2017,2018", NULL, 0},
  {"e-mail: _CTPAX_@MAIL.RU", NULL, 0},
  {"http://ctpax-cheater.losthost.org/htmldocs/ps_vs_ex.htm", NULL, 0},
  {"", NULL, 0},
  {"Special thanks to:", NULL, 0},
  {"", NULL, 0},
  {"Corvin", NULL, 0},
  {"http://www.r-t-c-m.com", NULL, 0},
  {"", NULL, 0},
  {"BlackDoomer", NULL, 0},
  {"https://cher-nov.github.io", NULL, 0},
  {"", NULL, 0},
  {"Compatibility fixes for DOS and Windows old games and software.", NULL, 0},
  {"Hiring? Offering a one-time job? Paid requests are always welcome!", NULL, 0},
  {"Contact for more details.", NULL, 0}
};

static menuitem mainexit[] = {
  {"Save settings?", NULL, 2 | MENU_DLGBOX},
  {"Yes", "Save current settings and exit program.", 0},
  {"No", "Do not save current settings, but exit program.", 0}
};

static menuitem notfound[] = {
  {"Files not found", NULL, 1 | MENU_NFOBOX},
  {"No supported game executables found!", NULL, 0}
};

int MenuProc(int item) {
int r;
  r = 1;
  switch (item) {
    /* hacks */
    case 1:
      DrawMenu(mainctrl, NULL);
      break;
    /* video resolutions */
    case 2:
      DrawMenu(mainmode, NULL);
      break;
    /* v1.5 modding */
    case 3:
      if (DrawMenu(mainwarn, NULL) == 2) {
        DrawMenu(mainmods, NULL);
      }
      break;
    /* v1.5 saved games editor */
    case 4:
      EditSavedGames();
      break;
    /* about */
    case 5:
      DrawMenu(maininfo, NULL);
      /* v1.5 notes */
      TextInfo();
      break;
    /* exit */
    case 6:
    case 7:
      r = 0;
      break;
  }
  return(r);
}

void LoadPatches(int n) {
WORD w, h, x, y;
int fl, i;
char *s;
BYTE b;
  /* load mask */
  ReadBuildMask(0, 0);
  fl = open(gamelist[n].name, O_RDONLY | O_BINARY);
  if (fl != -1) {
    for (i = 0; i < INFOSCOUNT; i++) {
      /* lazy check - only first byte */
      lseek(fl, GET_OFFS(datalist[infolist[i].idx][n].offs), SEEK_SET);
      read(fl, &b, 1);
      infolist[i].stat = (b == (datalist[infolist[i].idx][n].offs >> 24)) ? 0 : 1;
      /* menu selection */
      if (i < mainctrl[0].flag) {
        mainctrl[i + 1].flag |= infolist[i].stat ? ITEM_HILITE : 0;
      }
    }
    /* mask not loaded */
    if (!MaskStatus()) {
      infolist[SCR_IDX].stat = 0;
      mainmode[0].flag = 1 | MENU_NFOBOX;
      mainmode[0].item = "Custom resolution hack disabled";
      mainmode[0].hint = NULL;
      mainmode[1].flag = ITEM_HILITE;
      mainmode[1].item = "Status bar mask file not found.";
      mainmode[1].hint = NULL;
    }
    /* screen resolutions hack */
    if (infolist[SCR_IDX].stat) {
      lseek(fl, GET_OFFS(datalist[infolist[SCR_IDX].idx + 3][n].offs), SEEK_SET);
      read(fl, &w, 2);
      lseek(fl, GET_OFFS(datalist[infolist[SCR_IDX].idx + 5][n].offs), SEEK_SET);
      read(fl, &h, 2);
      infolist[SCR_IDX].stat = 0;
      for (i = 1; i < mainmode[0].flag; i++) {
        s = mainmode[i + 1].item;
        x = StrToInt(s);
        for (; *s && (*s != 'x'); s++);
        s += *s ? 1 : 0;
        y = StrToInt(s);
        if ((w == x) && (h == y)) {
          /* drop disabled flag */
          mainmode[1].flag ^= ITEM_HILITE;
          /* set new flag */
          mainmode[i + 1].flag |= ITEM_HILITE;
          /* found */
          infolist[SCR_IDX].stat = 1;
          break;
        }
      }
    }
    /* v1.5 modding tools hacks */
    for (i = 0; i < MODSLCOUNT; i++) {
      lseek(fl, GET_OFFS(modslist[modsinfo[i].idx][n].offs), SEEK_SET);
      read(fl, &b, 1);
      modsinfo[i].stat = (b == (modslist[modsinfo[i].idx][n].offs >> 24)) ? 0 : 1;
      mainmods[1 + i].flag |= modsinfo[i].stat ? ITEM_HILITE : 0;
    }
    mainmods[1 + MODSLCOUNT].flag |= (StrBlockState(fl, n) == 0xFF) ? 0 : ITEM_HILITE;
    close(fl);
  }
}

void SavePatches(int n) {
struct ftime ft;
int fl, i, j;
WORD w;
BYTE b;
char *s;
  fl = open(gamelist[n].name, O_RDWR | O_BINARY);
  if (fl != -1) {
    getftime(fl, &ft);
    /* v1.5 first of all rollback all superseded hacks from older versions */
    for (i = 0; i < OLD_COUNT; i++) {
      lseek(fl, GET_OFFS(old_list[i][n].offs), SEEK_SET);
      read(fl, &b, 1);
      /* pached byte? */
      if (b == old_list[i][n].data) {
        /* replace with original */
        lseek(fl, GET_OFFS(old_list[i][n].offs), SEEK_SET);
        b = (old_list[i][n].offs >> 24);
        write(fl, &b, 1);
      }
    }
    /* v1.5 fix snake cam cheat text if decryption modding hack used */
    b = StrBlockState(fl, n);
    if (b != 0xFF) {
      j = infolist[CHT_IDX].idx + infolist[CHT_IDX].count - CHT_LEN;
      for (i = 0; i < CHT_LEN; i++) {
        datalist[j + i][n].data ^= 0xFFUL;
        datalist[j + i][n].data ^= b;
        datalist[j + i][n].offs ^= 0xFF000000UL;
        datalist[j + i][n].offs ^= ((DWORD) b) << 24;
      }
    }
    /* screen resolutions hack */
    infolist[SCR_IDX].stat = (mainmode[1].flag & ITEM_HILITE) ? 0 : 1;
    if (infolist[SCR_IDX].stat) {
      infolist[SCR_IDX].stat = 0;
      for (i = 1; i < mainmode[0].flag; i++) {
        if (mainmode[i + 1].flag & ITEM_HILITE) {
          s = mainmode[i + 1].item;
          w = StrToInt(s);
          datalist[infolist[SCR_IDX].idx + 3][n].data = (w & 0xFF);
          datalist[infolist[SCR_IDX].idx + 4][n].data = ((w >> 8) & 0xFF);
          /* create mask file */
          ReadBuildMask(w, 1);
          for (; *s && (*s != 'x'); s++);
          s += *s ? 1 : 0;
          w = StrToInt(s);
          datalist[infolist[SCR_IDX].idx + 5][n].data = (w & 0xFF);
          datalist[infolist[SCR_IDX].idx + 6][n].data = ((w >> 8) & 0xFF);
          /* found */
          infolist[SCR_IDX].stat = 1;
          break;
        }
      }
    } else {
      /* delete mask file */
      ReadBuildMask(0, 2);
    }
    for (i = 0; i < INFOSCOUNT; i++) {
      if (i < mainctrl[0].flag) {
        infolist[i].stat = (mainctrl[i + 1].flag & ITEM_HILITE) ? 1 : 0;
      }
      for (j = 0; j < infolist[i].count; j++) {
        /* offset must exists */
        if (datalist[infolist[i].idx + j][n].offs) {
          lseek(fl, GET_OFFS(datalist[infolist[i].idx + j][n].offs), SEEK_SET);
          b = infolist[i].stat ? datalist[infolist[i].idx + j][n].data : (datalist[infolist[i].idx + j][n].offs >> 24);
          write(fl, &b, 1);
        }
      }
    }
    /* v1.5 sfx hack */
    BuildVoiceStub(infolist[VOC_IDX].stat);
    /* v1.5 modding hacks */
    StrBlockDecrypt(fl, n | ((mainmods[1 + MODSLCOUNT].flag & ITEM_HILITE) ? 2 : 0));
    for (i = 0; i < MODSLCOUNT; i++) {
      modsinfo[i].stat = (mainmods[1 + i].flag & ITEM_HILITE) ? 1 : 0;
      for (j = 0; j < modsinfo[i].count; j++) {
        lseek(fl, GET_OFFS(modslist[modsinfo[i].idx + j][n].offs), SEEK_SET);
        b = modsinfo[i].stat ? modslist[modsinfo[i].idx + j][n].data : (modslist[modsinfo[i].idx + j][n].offs >> 24);
        write(fl, &b, 1);
      }
    }
    /* v1.5 world map */
    j = 0;
    for (i = 0; i < 4; i++) {
      if (mainmods[1 + mainmods[0].flag - 4 + i].flag & ITEM_HILITE) {
        j = i;
        break;
      }
    }
    WorldMapDumper(fl, n | (j << 1));
    setftime(fl, &ft);
    close(fl);
  }
}

int main(void) {
char s[34];
int i, n;
  /* startup */
  InitSetup(maininfo[1].item, maininfo[2].item);
  /* find required files */
  maingame[0].item = "Game selection";
  maingame[0].hint = "Available files";
  maingame[0].flag = 0;
  for (i = 0; i < GAMESCOUNT; i++) {
    if (FileExists(gamelist[i].name, gamelist[i].size)) {
      maingame[0].flag++;
      maingame[maingame[0].flag].item = gamelist[i].name;
      maingame[maingame[0].flag].hint = gamelist[i].game;
      maingame[maingame[0].flag].flag = 0;
    }
  }
  /* if no game executables found */
  i = maingame[0].flag;
  if (!i) {
    /* mark as no files */
    i = -1;
    /* v1.4 disable hacks menu items */
    mainmenu[1].flag |= ITEM_GRAYED; /* disable hacks menu */
    mainmenu[2].flag |= ITEM_GRAYED; /* disable resolution menu */
    mainmenu[3].flag |= ITEM_GRAYED; /* disable modding menu */
    mainmenu[6].flag |= ITEM_GRAYED; /* disable save changes menu */
    /* show error message */
    DrawMenu(notfound, NULL);
  }
  /* multiple files found - select one */
  if (i > 1) {
    i = DrawMenu(maingame, NULL);
  }
  /* user actually selects something */
  if (i) {
    n = i;
    /* at least one file */
    if (n != -1) {
      /* generate menu subheader */
      StrMerge(s, 34, "Current Filename ( ", maingame[n].item, " )", NULL);
      mainmenu[0].hint = s;
      /* get index of the selected item */
      for (i = 0; i < GAMESCOUNT; i++) {
        if (MatchItem(gamelist[i].name, maingame[n].item)) {
          n = i;
          break;
        }
      }
      /* load patches */
      LoadPatches(n);
    }
    /* show main menu */
    i = DrawMenu(mainmenu, MenuProc);
    /* if save changes or yes in additional menu */
    if ((n != -1) && ((i == 6) || ((i == 0) && (DrawMenu(mainexit, NULL) == 1)))) {
      /* save patches */
      SavePatches(n);
    }
  }
  /* cleanup */
  FreeSetup();
  return(0);
}
