#ifndef __XSMODSVG_H
#define __XSMODSVG_H

#include <io.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "modsetup.h"
#include "xscommon.h"

/* items indexes */
#define EP_ITEM_HEART  0
#define EP_ITEM_SCARAB 1
#define EP_ITEM_TORCH  2
#define EP_ITEM_HAND   3
#define EP_ITEM_EYE    4
#define EP_ITEM_MASK   5

/* weapon order */
#define EP_WEAPON_MACHETE      0
#define EP_WEAPON_MAGNUM       1
#define EP_WEAPON_M60          2
#define EP_WEAPON_FLAMETHROWER 3
#define EP_WEAPON_GRENADES     4
#define EP_WEAPON_COBRASTAFF   5
#define EP_WEAPON_RAHGAUNTLET  6

/* weapon bit mask */
#define EP_WEAPON_MASK_BIT(x) (1<<(x))

/* max ingame ammo for each weapon */
#define EP_WEAPON_AMMO_MAX 300

/* only for M-60 since Magnum ammo will be at full capacity after load */
#define EP_WEAPON_CLIP_MAX 99

/* magic max */
#define EP_MAGIC_MAX 1000

/* max save slots */
#define EP_SAVES_MAX 5
/* save game length */
#define EP_SAVES_LEN 25

#pragma pack(push, 1)
#ifdef NEVER
struct {
  WORD  gFXVolume;    /* sound fx volume */
  WORD  gMusicVolume; /* music volume */
  WORD  nScreenSize;  /* screen size */
  WORD  bFullScreen;  /* 1 - HUD visible; 0 - no HUD (fullscreen) */
  WORD  nGamma;       /* gamma level */
  DWORD lMouseSens;   /* mouse sensitivity */
} psa_ini;
/* 14 bytes */
#endif

typedef struct {
  WORD BloodAmount;    /* ignored by game (reset to 800 in signle and to 1600 in multi-player) */
  WORD GameParams1[9]; /* various game parameters1 */
  WORD KeysBitMask;    /* ignored by game (reset to 0) */ 
  WORD MagicAmount;    /* can be changed (1000 max) */
  BYTE ItemsAmount[8]; /* last two unused */
  WORD AmmosAmount[9]; /* first (machete) and last two unused */
  WORD GameParams2[7]; /* various game parameters2 */
} TPlayerList;
/* 64 bytes */

typedef struct {
  BYTE CurrentLevel;      /* current level (0 - training; 1-20 - single; 21-32 - multi) */
  WORD PlayerWeapons;     /* available weapon bit mask */
  WORD PlayerWeapon;      /* current weapon */
  WORD PlayerClip;        /* M-60 cartridge belt ammo counter */
  WORD PlayerItem;        /* current inventory item */
  TPlayerList PlayerList; /* player list data */
  WORD PlayerLives;       /* player lives (5 max) */
} TGameStats;
/* 75 bytes */

typedef struct {
  char names[EP_SAVES_MAX][EP_SAVES_LEN]; /* if first char 0 - empty slot */
  TGameStats GameStats[EP_SAVES_MAX];
  BYTE Unused; /* unused (game write 15 for new file here but never checks) */
} TSavedGames;
#pragma pack(pop)
/* 501 */

void EditSavedGames(void);

#endif
